import asyncio
import time
import threading

from optscale_arcee.sender.sender import Sender
from optscale_arcee.name_generator import NameGenerator


def single(class_):
    instances = {}

    def get_instance(*args, **kwargs):
        if class_ not in instances:
            instances[class_] = class_(*args, **kwargs)
        return instances[class_]

    return get_instance


class Job(threading.Thread):
    def __init__(self, shutdown_flag, *args, **kwargs):
        # TODO: typing
        threading.Thread.__init__(self)
        self.__shutdown_flag = shutdown_flag
        self.__kw = kwargs

    def s_noblock(self, sender, run, token):
        return asyncio.run(sender.send_proc_data(run, token))

    def job(self):
        args = self.__kw.get("meth_args", list())
        self.s_noblock(*args)

    def run(self):
        sleep = self.__kw.get("sleep")
        if not sleep or not isinstance(sleep, int):
            # 1 second by default
            sleep = 1
        while not self.__shutdown_flag.is_set():
            self.job()
            time.sleep(sleep)


@single
class Arcee:
    def __init__(
            self, token=None, model_key=None, endpoint_url=None, ssl=True
    ):
        self.shutdown_flag = threading.Event()
        self.token = token
        self.model_key = model_key
        self.sender = Sender(endpoint_url, ssl, self.shutdown_flag)
        self.hb = None
        self._run = None
        self._tags = {}
        self._name = None

    @property
    def run(self):
        return self._run

    @run.setter
    def run(self, value):
        self._run = value

    @property
    def tags(self):
        return self._tags

    @tags.setter
    def tags(self, value):
        k, v = value
        self._tags.update({k: v})

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            finish()
        else:
            error()
        return exc_type is None


def init(token, model_key, run_name=None, endpoint_url=None, ssl=True, period=1):
    arcee = Arcee(token, model_key, endpoint_url, ssl)
    name = run_name if run_name is not None else NameGenerator.get_random_name()
    run_id = asyncio.run(arcee.sender.get_run_id(model_key, token, name))["id"]
    arcee.run = run_id
    arcee.name = name
    arcee.hb = Job(meth_args=(arcee.sender, run_id, token), sleep=period,
                   shutdown_flag=arcee.shutdown_flag)
    arcee.hb.start()
    asyncio.run(
        arcee.sender.send_stats(
            arcee.token,
            {"project": arcee.model_key, "run": arcee.run, "data": {}},
        )
    )
    return arcee


def tag(key, value):
    arcee = Arcee()
    arcee.tags = (key, value)
    asyncio.run(arcee.sender.add_tags(arcee.run, arcee.token, arcee.tags))


def milestone(value):
    arcee = Arcee()
    asyncio.run(arcee.sender.add_milestone(arcee.run, arcee.token, value))


def stage(name):
    arcee = Arcee()
    asyncio.run(arcee.sender.create_stage(arcee.run, arcee.token, name))


def finish():
    arcee = Arcee()
    try:
        asyncio.run(
            arcee.sender.change_state(
                arcee.run,
                arcee.token,
                2,
                True,
            )
        )
    finally:
        arcee.shutdown_flag.set()
        arcee.hb.join()


def error():
    arcee = Arcee()
    try:
        asyncio.run(
            arcee.sender.change_state(
                arcee.run,
                arcee.token,
                3,
                True,
            )
        )
    finally:
        arcee.shutdown_flag.set()
        arcee.hb.join()


def info():
    arcee = Arcee()
    return arcee.__dict__


def send(data):
    arcee = Arcee()
    asyncio.run(
        arcee.sender.send_stats(
            arcee.token,
            {"project": arcee.model_key, "run": arcee.run, "data": data},
        )
    )



import json
import copy


class TestDataAzure:
    @staticmethod
    def get_test_data():
        return json.dumps(
            {
                "compute": {
                    "azEnvironment": "AzurePublicCloud",
                    "customData": "",
                    "evictionPolicy": "",
                    "extendedLocation": {"name": "", "type": ""},
                    "isHostCompatibilityLayerVm": "false",
                    "licenseType": "",
                    "location": "westeurope",
                    "name": "vlad-ml-mini",
                    "offer": "0001-com-ubuntu-server-focal",
                    "osProfile": {
                        "adminUsername": "ubuntu",
                        "computerName": "vlad-ml-mini",
                        "disablePasswordAuthentication": "true",
                    },
                    "osType": "Linux",
                    "placementGroupId": "",
                    "plan": {"name": "", "product": "", "publisher": ""},
                    "platformFaultDomain": "0",
                    "platformSubFaultDomain": "",
                    "platformUpdateDomain": "0",
                    "priority": "",
                    "provider": "Microsoft.Compute",
                    "publicKeys": [
                        {
                            "keyData": "ssh-rsa AAAAB\n",
                            "path": "/home/ubuntu/.ssh/authorized_keys",
                        }
                    ],
                    "publisher": "canonical",
                    "resourceGroupName": "vlad-rg",
                    "resourceId": "/subscriptions/..",
                    "securityProfile": {
                        "secureBootEnabled": "false",
                        "virtualTpmEnabled": "false",
                    },
                    "sku": "20_04-lts-gen2",
                    "storageProfile": {
                        "dataDisks": [],
                        "imageReference": {
                            "id": "",
                            "offer": "0001-com-ubuntu-server-focal",
                            "publisher": "canonical",
                            "sku": "20_04-lts-gen2",
                            "version": "latest",
                        },
                        "osDisk": {
                            "caching": "ReadWrite",
                            "createOption": "FromImage",
                            "diffDiskSettings": {"option": ""},
                            "diskSizeGB": "30",
                            "encryptionSettings": {"enabled": "false"},
                            "image": {"uri": ""},
                            "managedDisk": {
                                "id": "/subscriptions/7ab2fde4b3b6df3300",
                                "storageAccountType": "Premium_LRS",
                            },
                            "name": "vlad-ml-mini_OsDisk_1_c63ece7f1e0340aab0",
                            "osType": "Linux",
                            "vhd": {"uri": ""},
                            "writeAcceleratorEnabled": "false",
                        },
                        "resourceDisk": {"size": "16384"},
                    },
                    "subscriptionId": "fff",
                    "tags": "",
                    "tagsList": [],
                    "userData": "",
                    "version": "20.04.202209050",
                    "virtualMachineScaleSet": {"id": ""},
                    "vmId": "7940abd0-a8c6-403f-8102-e47738e3cd4a",
                    "vmScaleSetName": "",
                    "vmSize": "Standard_D2s_v3",
                    "zone": "1",
                },
                "network": {
                    "interface": [
                        {
                            "ipv4": {
                                "ipAddress": [
                                    {
                                        "privateIpAddress": "172.18.0.5",
                                        "publicIpAddress": "",
                                    }
                                ],
                                "subnet": [{"address": "172.18.0.0", "prefix": "24"}],
                            },
                            "ipv6": {"ipAddress": []},
                            "macAddress": "000D3A46E876",
                        }
                    ]
                },
            }
        )

    @classmethod
    def get_test_data_no_network(cls):
        d = copy.copy(json.loads(cls.get_test_data()))
        d["network"] = dict()
        return json.dumps(d)



import aiohttp
import threading

from optscale_arcee.platform import CollectorFactory
from optscale_arcee.module_collector.collector import Collector as ImportsCollector
from optscale_arcee.hw_stat_collector.collector import Collector as HwCollector


def check_shutdown_flag_set(function):
    async def inner(self, *args, **kwargs):
        if not self.shutdown_flag.is_set():
            return await function(self, *args, **kwargs)
    return inner


class Sender:

    # default OptScale url
    base_url = "https://my.optscale.com:443/arcee/v2"

    def __init__(self, endpoint_url=None, ssl=True, shutdown_flag=None):
        if endpoint_url is None:
            endpoint_url = self.base_url
        self.endpoint_url = endpoint_url
        self.shutdown_flag = shutdown_flag or threading.Event()
        self.ssl = ssl

    @staticmethod
    async def m():
        platform = await CollectorFactory.get()
        return await platform().get_platform_meta()

    @staticmethod
    async def _proc_data():
        return await HwCollector.collect_stats()

    @staticmethod
    async def _imports_data():
        return await ImportsCollector.get_imports()

    async def send_get_request(self, url, headers=None, params=None) -> str:
        async with aiohttp.ClientSession(headers=headers) as session:
            async with session.get(
                url, params=params, raise_for_status=True, ssl=self.ssl
            ) as response:
                return await response.json()

    async def send_post_request(self, url, headers=None, data=None) -> str:
        async with aiohttp.ClientSession(headers=headers) as session:
            async with session.post(
                url, json=data, raise_for_status=True, ssl=self.ssl
            ) as response:
                return await response.json()

    async def send_patch_request(self, url, headers=None, data=None) -> str:
        async with aiohttp.ClientSession(headers=headers) as session:
            async with session.patch(
                url, json=data, raise_for_status=True, ssl=self.ssl
            ) as response:
                return await response.json()

    @check_shutdown_flag_set
    async def get_run_id(self, model_key, token, run_name):
        uri = "%s/applications/%s/run" % (self.endpoint_url, model_key)
        headers = {"x-api-key": token, "Content-Type": "application/json"}
        data = {"imports": await self._imports_data(), "name": run_name}
        return await self.send_post_request(uri, headers, data)

    @check_shutdown_flag_set
    async def add_milestone(self, run_id, token, value):
        uri = "%s/run/%s/milestones" % (self.endpoint_url, run_id)
        headers = {"x-api-key": token, "Content-Type": "application/json"}
        return await self.send_post_request(uri, headers, {"milestone": value})

    @check_shutdown_flag_set
    async def add_tags(self, run_id, token, tags):
        uri = "%s/run/%s" % (self.endpoint_url, run_id)
        headers = {"x-api-key": token, "Content-Type": "application/json"}
        return await self.send_patch_request(uri, headers, {"tags": tags})

    @check_shutdown_flag_set
    async def change_state(self, run_id, token, state, finish=False):
        uri = "%s/run/%s" % (self.endpoint_url, run_id)
        headers = {"x-api-key": token, "Content-Type": "application/json"}
        return await self.send_patch_request(
            uri, headers, {"state": state, "finish": finish}
        )

    @check_shutdown_flag_set
    async def create_stage(self, run_id, token, name):
        uri = "%s/run/%s/stages" % (self.endpoint_url, run_id)
        headers = {"x-api-key": token, "Content-Type": "application/json"}
        return await self.send_post_request(uri, headers, {"stage": name})

    @check_shutdown_flag_set
    async def send_stats(self, token, data):
        headers = {"x-api-key": token, "Content-Type": "application/json"}
        meta = await self.m()
        data.update({"platform": meta.to_dict()})
        await self.send_post_request(
            "%s/%s" % (self.endpoint_url, "collect"), headers, data
        )

    @check_shutdown_flag_set
    async def send_proc_data(self, run_id, token):
        uri = "%s/run/%s/proc" % (self.endpoint_url, run_id)
        headers = {"x-api-key": token, "Content-Type": "application/json"}
        data = dict()
        meta = await self.m()
        proc = await self._proc_data()
        data.update({"platform": meta.to_dict()})
        data.update({"proc_stats": proc})
        return await self.send_post_request(uri, headers, data)



import asyncio
import concurrent.futures
from functools import partial
import inspect
from modulefinder import ModuleFinder
import os
import sys
import __main__


class Collector:

    executor = concurrent.futures.ThreadPoolExecutor(max_workers=3)

    """
    Modules collector
    """

    __filter__ = [
        "tf",  # Tensorflow
        "torch",  # PyTorch
        "sklearn",  # Scikit
        "keras",  # Keras
        "mxnet",  # MXNet
        "numpy",  # NumPy
        "scipy",  # SciPy
        "sklearn",  # Scikit,
        "theano",  # Theano
        "pandas",  # Pandas
    ]

    @staticmethod
    async def get_cfp(real: bool = False) -> str:
        frame = inspect.stack()[1]
        p = frame[0].f_code.co_filename
        if real:
            return os.path.realpath(p)
        return p

    @classmethod
    async def apply_filter(cls, modules):
        result = list()
        for i in cls.__filter__:
            for j in modules:
                if i in j:
                    result.append(i)
                    break
        return result

    @classmethod
    async def run_async(cls, func, *args, loop=None, executor=None, **kwargs):
        if loop is None:
            loop = asyncio.get_event_loop()
        if executor is None:
            executor = cls.executor
        pfunc = partial(func, *args, **kwargs)
        return await loop.run_in_executor(executor, pfunc)

    @staticmethod
    def _fallback():
        for m in sys.modules.values():
            if m:
                yield m.__name__

    @classmethod
    async def indirect(cls, file=None, use_cfp=False, f_simplify=True):
        if file is None:
            if not use_cfp:
                file = __main__.__file__
            else:
                file = await cls.get_cfp()
        try:
            finder = ModuleFinder()
            await cls.run_async(finder.run_script, file)
            m = finder.modules.keys()
        except (AttributeError, RecursionError):
            # try to bypass 40350/84530 ns mod issue
            # pls see following for details

            # https://bugs.python.org/issue40350
            # https://github.com/python/cpython/issues/84530
            # https://github.com/python/cpython/pull/19917
            # https://github.com/python/cpython/pull/29196
            m = await cls.run_async(cls._fallback)
        if f_simplify:
            mods = await cls.apply_filter(m)
        else:
            mods = m
        for name in mods:
            await asyncio.sleep(0)
            yield name

    @classmethod
    async def get_imports(cls):
        return [i async for i in cls.indirect()]



from typing import List


class AzureComputeMeta:
    """
    AzureComputeMeta ...
    """

    def __init__(
        self,
        name: str,
        location: str,
        resourceId: str,
        vmSize: str,
        zone: str,
        subscriptionId: str,
        **kwargs
    ):
        self.name = name
        self.location = location
        self.resourceId = resourceId.lower()
        self.vmSize = vmSize
        self.zone = zone
        self.subscriptionId = subscriptionId


class AzureIpMeta:
    def __init__(self, privateIpAddress: str, publicIpAddress: str, **kwargs):
        self.privateIpAddress = privateIpAddress
        self.publicIpAddress = publicIpAddress


class AzureSubnetMeta:
    def __init__(self, address: str, prefix: str):
        self.address = address
        self.prefix = prefix


class AzureIPv4Meta:
    def __init__(
        self, ipAddress: List[AzureIpMeta], subnet: List[AzureSubnetMeta], **kwargs
    ):
        self.ipAddress = ipAddress
        self.subnet = subnet

    @classmethod
    def from_dict(cls, d):
        ipAddress = d.get("ipAddress")
        if ipAddress:
            d["ipAddress"] = [AzureIpMeta(**addr) for addr in ipAddress]
        subnet = d.get("subnet")
        if subnet:
            d["subnet"] = [AzureSubnetMeta(**sbn) for sbn in subnet]
        return cls(**d)


class AzureInterfaceMeta:
    def __init__(self, ipv4: AzureIPv4Meta, macAddress: str, **kwargs):
        self.ipv4 = ipv4
        self.macAddress = macAddress

    @classmethod
    def from_dict(cls, d):
        ipv4 = d.get("ipv4")
        if ipv4:
            d["ipv4"] = AzureIPv4Meta.from_dict(ipv4)
        return cls(**d)


class AzureNetworkMeta:
    def __init__(self, interface: List[AzureInterfaceMeta]):
        self.interface = interface

    @classmethod
    def from_dict(cls, d):
        interface = d.get("interface")
        if interface:
            d["interface"] = [AzureInterfaceMeta.from_dict(i) for i in interface]
        return cls(**d)


class AzureMeta:
    def __init__(self, compute: AzureComputeMeta, network: AzureNetworkMeta, **kwargs):
        self.compute = compute
        self.network = network

    @classmethod
    def from_dict(cls, d):
        compute = d.get("compute")
        if compute:
            d["compute"] = AzureComputeMeta(**compute)
        network = d.get("network")
        if network:
            d["network"] = AzureNetworkMeta.from_dict(network)
        return cls(**d)



import platform
import aiohttp
import asyncio
import aiofiles
import json
from enum import Enum

from optscale_arcee.platforms_meta.azure import AzureMeta


def serialise(self) -> dict:
    result = dict()
    for k, v in self.__dict__.items():
        if k is not None:
            if isinstance(v, Enum):
                v = v.value
            result[k] = v
    return result


class PlatformType(Enum):
    ali = "ali"
    aws = "aws"
    azure = "azure"
    gcp = "gcp"
    unknown = "unknown"


class InstanceLifeCycle(Enum):
    OnDemand = "OnDemand"
    Preemptible = "Preemptible"
    Spot = "Spot"
    Unknown = "Unknown"


class PlatformMeta:
    def __init__(
        self,
        platform_type,
        instance_id="",
        account_id="",
        local_ip="",
        public_ip="",
        instance_lc=InstanceLifeCycle.Unknown,
        instance_type="",
        instance_region="",
        availability_zone="",
    ):
        self.platform_type = platform_type
        self.instance_id = instance_id
        self.account_id = account_id
        self.local_ip = local_ip
        self.public_ip = public_ip
        self.instance_lc = instance_lc
        self.instance_type = instance_type
        self.instance_region = instance_region
        self.availability_zone = availability_zone

    def to_dict(self) -> dict:
        return json.loads(json.dumps(self, default=serialise))


class Platform:

    match = {
        "Amazon EC2": PlatformType.aws,
        "Google": PlatformType.gcp,
        "Microsoft Corporation": PlatformType.azure,
        "Alibaba Cloud": PlatformType.ali,
    }

    def __int__(self, *args, **kwargs):
        raise TypeError("Can be used only statically")

    @staticmethod
    def platform_name() -> str:
        return platform.node()

    @staticmethod
    async def get_platform_vendor() -> PlatformType:
        try:
            async with aiofiles.open("/sys/hypervisor/uuid", "r") as vendor:
                data = await vendor.read()
                if data.rstrip().lower().startswith("ec2"):
                    return PlatformType.aws
        except (OSError, IOError):
            pass
        return PlatformType.unknown

    @classmethod
    async def board_version(cls) -> PlatformType:
        try:
            async with aiofiles.open("/sys/class/dmi/id/board_vendor", "r") as board:
                data = await board.read()
                return cls.match.get(data.rstrip(), PlatformType.unknown)
        except (OSError, IOError):
            pass

    @classmethod
    async def sys_vendor(cls) -> PlatformType:
        try:
            async with aiofiles.open("/sys/class/dmi/id/sys_vendor", "r") as board:
                data = await board.read()
                return cls.match.get(data.rstrip(), PlatformType.unknown)
        except (OSError, IOError):
            pass

    @classmethod
    async def platform(cls) -> PlatformType:
        pl = await cls.get_platform_vendor()
        if pl == PlatformType.unknown:
            pl = await cls.board_version()
        if pl == PlatformType.unknown:
            pl = await cls.sys_vendor()
        return pl


class BaseCollector:
    @staticmethod
    async def send_request(url, headers=None, params=None, response="json") -> str:
        async with aiohttp.ClientSession(headers=headers) as session:
            async with session.get(url, params=params) as response:
                if response == "json":
                    resp = await response.json()
                else:
                    resp = await response.text()
                return resp


class AwsCollector(BaseCollector):

    base_url = "http://169.254.169.254/latest/meta-data/%s"

    async def get_instance_id(self):
        return await self.send_request(self.base_url % "instance-id", response="text")

    async def get_account_id(self):
        acc_info = await self.send_request(
            self.base_url % "identity-credentials/ec2/info", response="json"
        )
        return json.loads(acc_info).get("AccountId", "")

    async def get_local_ip(self):
        return await self.send_request(self.base_url % "local-ipv4", response="text")

    async def get_public_ip(self):
        return await self.send_request(self.base_url % "public-ipv4", response="text")

    async def get_life_cycle(self):
        match = {
            "spot": InstanceLifeCycle.Spot,
            "on-demand": InstanceLifeCycle.OnDemand,
        }
        lc_fut = await self.send_request(
            self.base_url % "instance-life-cycle", response="text"
        )
        return match.get(lc_fut.lower(), InstanceLifeCycle.Unknown)

    async def get_instance_type(self):
        return await self.send_request(self.base_url % "instance-type", response="text")

    async def get_az(self):
        return await self.send_request(
            self.base_url % "placement/availability-zone", response="text"
        )

    async def get_region(self):
        return await self.send_request(
            self.base_url % "placement/region", response="text"
        )

    async def get_platform_meta(self):
        return PlatformMeta(
            PlatformType.aws,
            await self.get_instance_id(),
            await self.get_account_id(),
            await self.get_local_ip(),
            await self.get_public_ip(),
            await self.get_life_cycle(),
            await self.get_instance_type(),
            await self.get_region(),
            await self.get_az(),
        )


class GcpCollector(BaseCollector):

    base_url = "http://169.254.169.254/computeMetadata/v1/%s"
    headers = {"Metadata-Flavor": "Google"}

    async def get_instance_id(self):
        return await self.send_request(
            self.base_url % "instance/id", headers=self.headers, response="text"
        )

    async def get_account_id(self):
        return await self.send_request(
            self.base_url % "project/project-id", headers=self.headers, response="text"
        )

    async def get_local_ip(self):
        return await self.send_request(
            self.base_url % "instance/network-interfaces/0/ip",
            headers=self.headers,
            response="text",
        )

    async def get_public_ip(self):
        return await self.send_request(
            self.base_url
            % "instance/network-interfaces/0/access-configs/0/external-ip",
            headers=self.headers,
            response="text",
        )

    async def get_life_cycle(self):
        match = {
            "true": InstanceLifeCycle.Spot,
            "false": InstanceLifeCycle.OnDemand,
        }
        lc_fut = await self.send_request(
            self.base_url % "instance/scheduling/preemptible",
            headers=self.headers,
            response="text",
        )
        return match.get(lc_fut.lower(), InstanceLifeCycle.Unknown)

    async def get_instance_type(self):
        it_fut = await self.send_request(
            self.base_url % "instance/machine-type",
            headers=self.headers,
            response="text",
        )
        return it_fut.split("/")[-1]

    async def get_az(self):
        az_fut = await self.send_request(
            self.base_url % "placement/availability-zone",
            headers=self.headers,
            response="text",
        )
        return az_fut.split("/")[-1]

    async def get_region(self):
        # TODO
        await asyncio.sleep(0)
        return ""

    async def get_platform_meta(self):
        return PlatformMeta(
            PlatformType.gcp,
            await self.get_instance_id(),
            await self.get_account_id(),
            await self.get_local_ip(),
            await self.get_public_ip(),
            await self.get_life_cycle(),
            await self.get_instance_type(),
            await self.get_region(),
            await self.get_az(),
        )


class AzureCollector(BaseCollector):

    base_url = "http://169.254.169.254/metadata/instance/"

    async def collect(self) -> str:
        """
        Collects data from metadata server
        :return: (str) metadata json
        """
        headers = {"Metadata": "true"}
        params = [("format", "json"), ("api-version", "2021-10-01")]
        return await self.send_request(self.base_url, headers, params)

    async def get_platform_meta(self) -> PlatformMeta:
        meta_response = await self.collect()
        if meta_response:
            meta = AzureMeta.from_dict(json.loads(meta_response))
            private_ip = public_ip = ""
            # TODO: check usage
            if meta.network and meta.network.interface:
                private_ip = (
                    meta.network.interface[0].ipv4.ipAddress[0].privateIpAddress
                )
                public_ip = meta.network.interface[0].ipv4.ipAddress[0].publicIpAddress
            return PlatformMeta(
                PlatformType.azure,
                meta.compute.resourceId,
                meta.compute.subscriptionId,
                private_ip,
                public_ip,
                InstanceLifeCycle.Unknown,
                meta.compute.vmSize,
                meta.compute.location,
                meta.compute.zone,
            )
        return PlatformMeta(PlatformType.azure)


class UnknownCollector(BaseCollector):
    @staticmethod
    async def send_request(url, headers=None, params=None, response="json") -> str:
        return await asyncio.sleep(0)

    @staticmethod
    async def get_platform_meta() -> PlatformMeta:
        return PlatformMeta(PlatformType.unknown, Platform.platform_name())


class CollectorFactory:

    match = {
        PlatformType.azure: AzureCollector,
        PlatformType.aws: AwsCollector,
    }

    @classmethod
    async def get(cls):
        pf = await Platform.platform()
        return cls.match.get(pf, UnknownCollector)
